import { RecursionResult } from './RecursionResult';
export interface RecursionBuilderState<K, V, O extends object> {
    yieldOn?(value: V, key: K, object: O): boolean;
    traverseOn?(value: V, key: K, object: O): boolean;
    readEntry?(object: O): Iterable<[K, V]>;
    readNext?(object: V | O): V | O;
}
/**
 * Creates an iterable for recursively accessing an object heirarchy.
 * @export
 * @class RecursionBuilder
 * @implements {Iterable<[ K, V, string, O ]>}
 * @template K The type of the key.
 * @template V The type of the value.
 * @template O The type of the data object.
 */
export declare class RecursionBuilder<K = string, V = any, O extends object = object> implements Iterable<RecursionResult<K, V, O>> {
    private _object;
    private _yieldOn;
    private _traverseOn;
    private _readEntry;
    private _readNext?;
    /**
     * Creates an instance of RecursionBuilder.
     * @param {O} [_object]
     * @param {RecursionBuilderState<K, V, O>} [state={}]
     */
    constructor(_object?: O | undefined, state?: RecursionBuilderState<K, V, O>);
    /**
     * Sets the yield filter and returns a new RecursionBuilder. Determines whether
     * to yield the value.
     * @param {(value: V, key: K, object: O) => boolean} fn
     * @returns {RecursionBuilder<K, V, O>}
     */
    yieldOn(fn: (value: V, key: K, object: O) => boolean): RecursionBuilder<K, V, O>;
    /**
     * Sets the traverse filter and returns a new RecursionBuilder. Determines whether
     * to traverse the value given.
     * @param {(value: V, key: K, object: O) => boolean} fn
     * @returns {RecursionBuilder<K, V, O>}
     */
    traverseOn(fn: (value: V, key: K, object: O) => boolean): RecursionBuilder<K, V, O>;
    /**
     * Sets the read entry function and returns a new RecursionBuilder. This function
     * generates the key/value pairs of each key.
     * @param {(object: O) => Iterable<[ K, V ]>} fn
     * @returns {RecursionBuilder<K, V, O>}
     */
    readEntry(fn: (object: O) => Iterable<[K, V]>): RecursionBuilder<K, V, O>;
    /**
     * Sets the read next function and returns a new RecursionBuilder. This function
     * determines what value to read next. Useful for iterarting recursive arrays.
     *
     * @param {(((object: O) => O)|string)} fn
     * @returns {RecursionBuilder<K, V, O>}
     */
    readNext(fn: ((object: O) => O) | string): RecursionBuilder<K, V, O>;
    /**
     * Recurses the data object using the configured recursion algorithm.
     * @param {O} [object]
     * @param {string} [_path]
     * @returns {IterableIterator<[ K, V, string, O ]>}
     */
    recurse(object?: O): IterableIterator<RecursionResult<K, V, O>>;
    /**
     * Clones this recursion builder instance.
     * @param {RecursionBuilderState<K, V, O>} [state={}]
     * @returns {RecursionBuilder<K, V, O>}
     */
    clone(state?: RecursionBuilderState<K, V, O>): RecursionBuilder<K, V, O>;
    [Symbol.iterator](): IterableIterator<RecursionResult<K, V, O>>;
    /**
     * Iterates and returns only value results.
     * @param {O} [object]
     * @returns {IterableIterator<V>}
     */
    values(object?: O): IterableIterator<V>;
    /**
     * Iterates and returns only keys.
     * @param {O} [object]
     * @returns {IterableIterator<K>}
     */
    keys(object?: O): IterableIterator<K>;
    /**
     * Iterates and returns only paths.
     * @param {O} [object]
     * @returns {IterableIterator<string>}
     */
    paths(object?: O): IterableIterator<string>;
    /**
     * Iterates and returns only parent objects.
     * @param {O} [object]
     * @returns {IterableIterator<string>}
     */
    parents(object?: O): IterableIterator<O>;
    private _extractProp(object, key);
    /**
     * Recurses the data object using the configured recursion algorithm.
     * @param {O} [object]
     * @param {string} [_path]
     * @returns {IterableIterator<[ K, V, string, O ]>}
     */
    private _recurse(object, _path?, lastResult?);
    static create<K, V, O extends object>(object?: O, state?: RecursionBuilderState<K, V, O>): RecursionBuilder<K, V, O>;
}
