"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const RecursionResult_1 = require("./RecursionResult");
const isObject = (v) => typeof v === 'object' && v != null;
/**
 * Creates an iterable for recursively accessing an object heirarchy.
 * @export
 * @class RecursionBuilder
 * @implements {Iterable<[ K, V, string, O ]>}
 * @template K The type of the key.
 * @template V The type of the value.
 * @template O The type of the data object.
 */
class RecursionBuilder {
    /**
     * Creates an instance of RecursionBuilder.
     * @param {O} [_object]
     * @param {RecursionBuilderState<K, V, O>} [state={}]
     */
    constructor(_object, state = {}) {
        this._object = _object;
        this._yieldOn = state.yieldOn || (() => true);
        this._traverseOn = state.traverseOn || isObject;
        this._readEntry = state.readEntry || objectEntryReader;
        this._readNext = state.readNext;
    }
    /**
     * Sets the yield filter and returns a new RecursionBuilder. Determines whether
     * to yield the value.
     * @param {(value: V, key: K, object: O) => boolean} fn
     * @returns {RecursionBuilder<K, V, O>}
     */
    yieldOn(fn) {
        return this.clone({ yieldOn: fn });
    }
    /**
     * Sets the traverse filter and returns a new RecursionBuilder. Determines whether
     * to traverse the value given.
     * @param {(value: V, key: K, object: O) => boolean} fn
     * @returns {RecursionBuilder<K, V, O>}
     */
    traverseOn(fn) {
        return this.clone({ traverseOn: fn });
    }
    /**
     * Sets the read entry function and returns a new RecursionBuilder. This function
     * generates the key/value pairs of each key.
     * @param {(object: O) => Iterable<[ K, V ]>} fn
     * @returns {RecursionBuilder<K, V, O>}
     */
    readEntry(fn) {
        return this.clone({ readEntry: fn });
    }
    /**
     * Sets the read next function and returns a new RecursionBuilder. This function
     * determines what value to read next. Useful for iterarting recursive arrays.
     *
     * @param {(((object: O) => O)|string)} fn
     * @returns {RecursionBuilder<K, V, O>}
     */
    readNext(fn) {
        return this.clone({ readNext: createChildExtractor(fn) });
    }
    /**
     * Recurses the data object using the configured recursion algorithm.
     * @param {O} [object]
     * @param {string} [_path]
     * @returns {IterableIterator<[ K, V, string, O ]>}
     */
    *recurse(object) {
        if (!object) {
            return;
        }
        yield* this._recurse(object);
    }
    /**
     * Clones this recursion builder instance.
     * @param {RecursionBuilderState<K, V, O>} [state={}]
     * @returns {RecursionBuilder<K, V, O>}
     */
    clone(state = {}) {
        return RecursionBuilder.create(this._object, Object.assign({ yieldOn: this._yieldOn, traverseOn: this._traverseOn, readEntry: this._readEntry, readNext: this._readNext }, state));
    }
    *[Symbol.iterator]() {
        yield* this.recurse(this._object);
    }
    /**
     * Iterates and returns only value results.
     * @param {O} [object]
     * @returns {IterableIterator<V>}
     */
    *values(object) {
        yield* this._extractProp(object, 'value');
    }
    /**
     * Iterates and returns only keys.
     * @param {O} [object]
     * @returns {IterableIterator<K>}
     */
    *keys(object) {
        yield* this._extractProp(object, 'key');
    }
    /**
     * Iterates and returns only paths.
     * @param {O} [object]
     * @returns {IterableIterator<string>}
     */
    *paths(object) {
        yield* this._extractProp(object, 'path');
    }
    /**
     * Iterates and returns only parent objects.
     * @param {O} [object]
     * @returns {IterableIterator<string>}
     */
    *parents(object) {
        yield* this._extractProp(object, 'parents');
    }
    *_extractProp(object, key) {
        for (const results of this.recurse(object || this._object)) {
            yield results[key];
        }
    }
    /**
     * Recurses the data object using the configured recursion algorithm.
     * @param {O} [object]
     * @param {string} [_path]
     * @returns {IterableIterator<[ K, V, string, O ]>}
     */
    *_recurse(object, _path, lastResult) {
        for (const [key, value] of this._readEntry(object)) {
            const path = resolvePath(_path, key, object);
            let nextValue = value;
            let result = lastResult ? lastResult : undefined;
            if (this._yieldOn(value, key, object)) {
                result = new RecursionResult_1.RecursionResult(value, key, path, object, lastResult ? lastResult : null);
                yield result;
            }
            if (this._readNext) {
                nextValue = this._readNext(value);
            }
            if (this._traverseOn(nextValue, key, object)) {
                yield* this._recurse(nextValue, path, result);
            }
        }
    }
    static create(object, state) {
        return new RecursionBuilder(object, state);
    }
}
exports.RecursionBuilder = RecursionBuilder;
function createChildExtractor(fn) {
    return (value) => {
        return typeof fn === 'string' ? isObject(value) ? value[fn] : undefined : fn(value);
    };
}
function resolvePath(path, key, value) {
    return path ? Array.isArray(value) ? `${path}[${key}]` : `${path}.${key}` : key.toString();
}
function* objectEntryReader(object) {
    if (typeof object['entries'] === 'function') {
        yield* object['entries']();
    }
    else {
        for (const key of Object.keys(object)) {
            yield [key, object[key]];
        }
    }
}
//# sourceMappingURL=RecursionBuilder.js.map